﻿using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter {
  using System;
  using System.Collections;
  using System.Reflection;

  using Attributes;

  using Stack;

  public static class Extensions {
    private static readonly IDictionary<StackTypes, PropertyInfo> StackProperties = typeof(PushInterpreter)
        .GetProperties()
        .Select(p => new { Property = p, Attribute = p.GetCustomAttribute<PushStackAttribute>() })
        .Where(x => x.Attribute != null)
        .ToDictionary(x => x.Attribute.StackType, x => x.Property);

    public static IEnumerable<T> GetStackEntriesByType<T>(this IPushInterpreter interpreter, StackTypes stackType) {
      return StackProperties.ContainsKey(stackType)
        ? ((IEnumerable)StackProperties[stackType].GetValue(interpreter)).OfType<T>()
        : null;
    }

    public static Type GetStackEntryType(this StackTypes stackType) {
      return StackProperties.ContainsKey(stackType)
        ? StackProperties[stackType].PropertyType.GetGenericArguments().Single()
        : null;
    }

    public static void PrintStacks(this IPushInterpreter interpreter) {
      var stackTypesType = typeof(StackTypes);

      foreach (StackTypes type in Enum.GetValues(stackTypesType)) {
        if (!interpreter.Stacks.ContainsKey(type)) continue;

        var stackName = Enum.GetName(stackTypesType, type);
        var stack = interpreter.Stacks[type];

        if (stack.IsEmpty || !stack.IsEnabled) continue;

        var stackString = string.Join(" ", interpreter.StringifyStack(type).Reverse());
        Console.WriteLine("--------- {0} ---------\n{1}\n", stackName, stackString);
      }

      if (interpreter.CustomExpressions.Count > 0) {
        Console.WriteLine("--------- Custom Expressions ---------");
        foreach (var ce in interpreter.CustomExpressions) {
          Console.WriteLine("{0}: {1}", ce.Key, ce.Value);
        }
      }
    }
  }
}

